"""The intermediate query result used by the various ``build_...()`` methods."""

from __future__ import annotations

import logging
import operator
from datetime import date, datetime
from decimal import Decimal as D
from functools import reduce

from django.contrib.gis.geos import GEOSGeometry
from django.core.exceptions import FieldError
from django.db.models import Q, QuerySet
from django.db.models.expressions import Combinable, Func

from gisserver.features import FeatureType

logger = logging.getLogger(__name__)
ScalarTypes = bool | int | str | D | date | datetime
RhsTypes = Combinable | Func | Q | GEOSGeometry | bool | int | D | str | date | datetime | tuple


class CompiledQuery:
    """Intermediate data for translating FES queries to Django.

    This class effectively contains all data from the ``<fes:Filter>`` object,
    but using a format that can be translated to a Django QuerySet.

    As the Abstract Syntax Tree of a FES-filter creates the ORM query,
    it fills this object with all intermediate bits. This allows building
    the final QuerySet object in a single round. Each ``build_...()`` method
    in the tree may add extra lookups and annotations.
    """

    def __init__(
        self,
        feature_types: list[FeatureType],
        lookups: list[Q] | None = None,
        typed_lookups: dict[str, list[Q]] | None = None,
        annotations: dict[str, Combinable | Q] | None = None,
    ):
        """
        :param feature_types: The feature types this query uses.
                              Typically, this is one feature unless a JOIN syntax is used.

        The extra parameters of the init method ar typically used only in unit tests.
        """
        self.feature_types = feature_types
        self.lookups = lookups or []
        self.typed_lookups = typed_lookups or {}
        self.annotations = annotations or {}
        self.aliases = 0
        self.extra_lookups: list[Q] = []
        self.ordering: list[str] = []
        self.is_empty = False
        self.distinct = False

    def add_annotation(self, value: Combinable | Q) -> str:
        """Create a named-alias for a function/Q object.
        This alias can be used in a comparison, where expressions are used as left-hand-side.
        """
        self.aliases += 1
        name = f"a{self.aliases}"
        self.annotations[name] = value
        return name

    def add_distinct(self):
        """Enforce "SELECT DISTINCT" on the query, used when joining 1-N or N-M relationships."""
        self.distinct = True

    def add_lookups(self, q_object: Q, type_name: str | None = None):
        """Register an extra 'WHERE' clause of the query.
        This is used for comparisons, ID selectors and other query types.
        """
        if not isinstance(q_object, Q):
            raise TypeError()

        if type_name is not None:
            if type_name not in self.typed_lookups:
                self.typed_lookups[type_name] = []
            self.typed_lookups[type_name].append(q_object)
        else:
            self.lookups.append(q_object)

    def add_extra_lookup(self, q_object: Q):
        """Temporary stash an extra lookup that the expression can't return yet.
        This is used for XPath selectors that also filter on attributes,
        e.g. "element[@attr=..]/child". The attribute lookup is processed as another filter.
        """
        if not isinstance(q_object, Q):
            raise TypeError()
        # Note the ORM also provides a FilteredRelation() option, that is not explored yet here.
        self.extra_lookups.append(q_object)

    def add_ordering(self, ordering: list[str]):
        """Read the desired result ordering from a ``<fes:SortBy>`` element."""
        self.ordering.extend(ordering)

    def apply_extra_lookups(self, comparison: Q) -> Q:
        """Combine stashed lookups with the provided Q object.

        This is called for functions that compile a "Q" object.
        In case a node added extra lookups (for attributes), these are combined here
        with the actual comparison.
        """
        if not self.extra_lookups:
            return comparison

        # The extra lookups are used for XPath queries such as "/node[@attr=..]/foo".
        # A <ValueReference> with such lookup also requires to limit the filtered results,
        # in addition to the comparison operator code that is wrapped up here.
        result = reduce(operator.and_, [comparison] + self.extra_lookups)
        self.extra_lookups.clear()
        return result

    def mark_empty(self):
        """Mark as returning no results."""
        self.is_empty = True

    def get_queryset(self) -> QuerySet:
        """Apply the filters and lookups to the queryset."""
        queryset = self.feature_types[0].get_queryset()
        if self.is_empty:
            return queryset.none()

        if self.extra_lookups:
            # Each time an expression node calls add_extra_lookup(),
            # the parent should have used apply_extra_lookups()
            raise RuntimeError("apply_extra_lookups() was not called")

        # All are applied at once.
        if self.annotations:
            queryset = queryset.annotate(**self.annotations)

        lookups = self.lookups
        try:
            lookups += self.typed_lookups.pop(self.feature_types[0].xml_name)
        except KeyError:
            pass
        if self.typed_lookups:
            raise RuntimeError(
                "Types lookups defined for unknown feature types: %r", list(self.typed_lookups)
            )

        if lookups:
            try:
                queryset = queryset.filter(*lookups)
            except FieldError as e:
                logger.debug("Query failed: %s, constructed query: %r", e.args[0], lookups)
                e.args = (f"{e.args[0]} Constructed query: {lookups!r}",) + e.args[1:]
                raise

        if self.ordering:
            queryset = queryset.order_by(*self.ordering)

        if self.distinct:
            queryset = queryset.distinct()

        return queryset

    def __repr__(self):
        return (
            "<CompiledQuery"
            f" annotations={self.annotations!r},"
            f" lookups={self.lookups!r},"
            f" typed_lookups={self.typed_lookups!r}>"
        )

    def __eq__(self, other):
        """For pytest comparisons."""
        if isinstance(other, CompiledQuery):
            return (
                other.lookups == self.lookups
                and other.typed_lookups == self.typed_lookups
                and other.annotations == self.annotations
            )
        else:
            return NotImplemented



"""Parsing and processing of the OGC Filter Encoding Standard 2.0 (FES).

This parses the XML syntax, and translates that into a Django ORM query.

It's also possible to register custom functions,
which will include Django ORM function as part of the filter language.

The full spec can be found at: https://www.ogc.org/publications/standard/filter/.
Secondly, using https://www.mediamaps.ch/ogc/schemas-xsdoc/sld/1.2/filter_xsd.html
can be very helpful to see which options each object type should support.
"""

from __future__ import annotations

from . import lookups  # noqa: F401 (need to register ORM lookups)
from .expressions import (
    BinaryOperator,
    BinaryOperatorType,
    Expression,
    Function,
    Literal,
    ValueReference,
)
from .filters import Filter
from .identifiers import Id, ResourceId, VersionActionTokens
from .operators import (
    BetweenComparisonOperator,
    BinaryComparisonName,
    BinaryComparisonOperator,
    BinaryLogicOperator,
    BinaryLogicType,
    BinarySpatialOperator,
    ComparisonOperator,
    DistanceOperator,
    DistanceOperatorName,
    ExtensionOperator,
    IdOperator,
    LikeOperator,
    LogicalOperator,
    MatchAction,
    Measure,
    NilOperator,
    NonIdOperator,
    NullOperator,
    Operator,
    SpatialOperator,
    SpatialOperatorName,
    TemporalOperator,
    TemporalOperatorName,
    UnaryLogicOperator,
    UnaryLogicType,
)
from .sorting import SortBy, SortOrder, SortProperty

__all__ = [
    "Filter",
    # Expressions
    "Expression",
    "Function",
    "Literal",
    "ValueReference",
    # WFS 1.0 expressions
    "BinaryOperator",
    "BinaryOperatorType",
    # Identifiers
    "Id",
    "ResourceId",
    "VersionActionTokens",
    # Operators
    "BetweenComparisonOperator",
    "BinaryComparisonName",
    "BinaryComparisonOperator",
    "BinaryLogicOperator",
    "BinaryLogicType",
    "BinarySpatialOperator",
    "ComparisonOperator",
    "DistanceOperator",
    "DistanceOperatorName",
    "ExtensionOperator",
    "IdOperator",
    "LikeOperator",
    "LogicalOperator",
    "MatchAction",
    "Measure",
    "NilOperator",
    "NonIdOperator",
    "NullOperator",
    "Operator",
    "SpatialOperator",
    "SpatialOperatorName",
    "TemporalOperator",
    "TemporalOperatorName",
    "UnaryLogicOperator",
    "UnaryLogicType",
    # Sorting
    "SortBy",
    "SortOrder",
    "SortProperty",
    "SortBy",
]



"""Entry point to handle queries.

WFS defines 2 query types:
- Adhoc queries are constructed directly from request parameters.
- Stored queries are defined first, and executed later.

Both use the FES (Filter Encoding Syntax) filtering logic internally.

The objects in this module closely follow the WFS spec.
By using the same type definitions, a lot of code logic follows naturally.
"""

from __future__ import annotations

from typing import ClassVar

from django.db.models import Q, QuerySet

from gisserver.exceptions import (
    ExternalValueError,
    InvalidParameterValue,
    OperationNotSupported,
)
from gisserver.features import FeatureType
from gisserver.parsers import fes20, wfs20
from gisserver.parsers.ast import AstNode
from gisserver.parsers.query import CompiledQuery
from gisserver.projection import FeatureProjection


class QueryExpression(AstNode):
    """WFS base class for all queries.
    This object type is defined in the WFS spec (as ``<fes:AbstractQueryExpression>``).

    The WFS server can initiate queries in multiple ways.
    This class provides the common interface for all these query types;
    whether the request provided "ad-hoc" parameters or called a stored procedure.
    Each query type has its own way of generating the actual database statement to perform.

    The subclasses can override the following logic:

    * :meth:`get_type_names` defines which types this query applies to.
    * :meth:`build_query` defines how to filter the queryset.

    For full control, these methods can also be overwritten instead:

    * :meth:`get_queryset` defines the full results.
    """

    #: Configuration for the 'locator' argument in exceptions
    query_locator: ClassVar[str] = None

    # QueryExpression
    #: The 'handle' that will be returned in exceptions.
    handle: str = ""

    #: Projection parameters (overwritten by subclasses)
    #: In the WFS spec, this is only part of the operation/presentation.
    #: For Django, we'd like to make this part of the query too.
    property_names: list[wfs20.PropertyName] | None = None  # PropertyName

    #: The valueReference for the GetPropertyValue call, provided here for extra ORM filtering.
    value_reference: fes20.ValueReference | None = None

    def bind(
        self,
        feature_types: list[FeatureType],
        value_reference: fes20.ValueReference | None = None,
    ):
        """Bind the query to context outside its tag definition.

        :param feature_types: The corresponding feature types for :meth:`get_type_names`.
        :param value_reference: Which field is returned (by ``GetPropertyValue``)
        """
        # Store the resolved feature types
        self.feature_types = feature_types

        # for GetPropertyValue
        if value_reference is not None:
            self.value_reference = value_reference

    def get_queryset(self) -> QuerySet:
        """Generate the queryset for the specific feature type.

        This method can be overwritten in subclasses to define the returned data.
        However, consider overwriting :meth:`build_query` instead of simple data.
        """
        if len(self.feature_types) > 1:
            raise OperationNotSupported("Join queries are not supported", locator="typeNames")

        # To apply the filters, an internal CompiledQuery object is created.
        # This collects all steps, to create the final QuerySet object.
        # The build_query() method may return an Q object, or fill the compiler itself.
        compiler = CompiledQuery(self.feature_types)
        q_object = self.build_query(compiler)
        if q_object is not None:
            compiler.add_lookups(q_object)

        # While property names are projection, this hook should
        # make it possible to perform extra query adjustments for complex expressions.
        if self.property_names:
            for property_name in self.property_names:
                property_name.decorate_query(compiler)

        if self.value_reference is not None:
            try:
                self.value_reference.parse_xpath(self.feature_types)
            except ExternalValueError as e:
                raise InvalidParameterValue(
                    f"Field '{self.value_reference.xpath}' does not exist.",
                    locator="valueReference",
                ) from e

            # For GetPropertyValue, adjust the query so only that value is requested.
            # This makes sure XPath attribute selectors are already handled by the
            # database query, instead of being a presentation-layer handling.
            # This supports cases like: ``addresses/Address[street="Oxfordstrasse"]/number``
            field = self.value_reference.build_rhs(compiler)

            queryset = compiler.get_queryset()
            return queryset.values("pk", member=field)
        else:
            return compiler.get_queryset()

    def get_type_names(self) -> list[str]:
        """Tell which type names this query applies to.
        Multiple values means a JOIN is made (not supported yet).

        This method needs to be defined in subclasses.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__}.get_type_names() should be implemented."
        )

    def get_projection(self) -> FeatureProjection:
        """Tell how the query should be displayed."""
        raise NotImplementedError()

    def build_query(self, compiler: CompiledQuery) -> Q | None:
        """Define the compiled query that filters the queryset."""
        raise NotImplementedError()

    def as_kvp(self) -> dict:
        """Translate the POST request into KVP GET parameters. This is needed for pagination."""
        params = {}
        if self.property_names:
            params["PROPERTYNAME"] = ",".join(p.xpath for p in self.property_names)
        return params



"""GML support for the fes filtering.

Overview of GML 3.2 changes: https://mapserver.org/el/development/rfc/ms-rfc-105.html#rfc105
"""

from dataclasses import dataclass
from xml.etree.ElementTree import tostring

from django.contrib.gis.gdal import AxisOrder
from django.contrib.gis.geos import GEOSGeometry, Polygon

from gisserver.crs import CRS
from gisserver.exceptions import ExternalParsingError
from gisserver.parsers.ast import tag_registry
from gisserver.parsers.query import CompiledQuery
from gisserver.parsers.xml import NSElement, xmlns

from .base import AbstractGeometry, AbstractTimePrimitive

_ANY_GML_NS = "{http://www.opengis.net/gml/"


def is_gml_element(element) -> bool:
    """Tell whether the element is an GML element."""
    return element.tag.startswith(_ANY_GML_NS)


@dataclass(repr=False)
@tag_registry.register("Polygon", xmlns.gml21)
@tag_registry.register("LineString", xmlns.gml32)
@tag_registry.register("LinearRing", xmlns.gml32)
@tag_registry.register("MultiLineString", xmlns.gml32)
@tag_registry.register("MultiPoint", xmlns.gml32)
@tag_registry.register("MultiPolygon", xmlns.gml32)
@tag_registry.register("MultiSurface", xmlns.gml32)
@tag_registry.register("Point", xmlns.gml32)
@tag_registry.register("Polygon", xmlns.gml32)
@tag_registry.register("Envelope", xmlns.gml32)
class GEOSGMLGeometry(AbstractGeometry):
    """Convert the incoming GML into a Django GEOSGeometry.
    This tag parses all ``<gml:...>`` geometry elements within the query.
    """

    # Not implemented:
    # - Curve
    # - MultiCurve
    # - MultiGeometry
    # - Surface

    srs: CRS
    geos_data: GEOSGeometry

    @classmethod
    def from_bbox(cls, bbox_value: str):
        """Parse the bounding box from an input string.

        It can either be 4 coordinates, or 4 coordinates with a special reference system.
        """
        bbox = bbox_value.split(",")
        if not (4 <= len(bbox) <= 5):
            raise ExternalParsingError(
                f"Input does not contain bounding box, expected 4 or 5 values, not {bbox}."
            )

        polygon = Polygon.from_bbox(
            (float(bbox[0]), float(bbox[1]), float(bbox[2]), float(bbox[3]))
        )
        if len(bbox) == 5:
            crs = CRS.from_string(bbox[4])
            polygon.srid = crs.srid
        else:
            crs = None  # will be resolved

        # Wrap in an element that the filter can use.
        CRS.tag_geometry(polygon, axis_order=AxisOrder.AUTHORITY)
        return cls(srs=crs, geos_data=polygon)

    @classmethod
    def from_xml(cls, element: NSElement):
        """Push the whole <gml:...> element into the GEOS parser.
        This avoids having to support the whole GEOS logic.

        GML is a complex beast with many different forms for the same thing:
        http://erouault.blogspot.com/2014/04/gml-madness.html
        """
        srs = CRS.from_string(element.get_str_attribute("srsName"))

        # Push the whole <gml:...> element into the GEOS parser.
        # This avoids having to support the whole GEOS logic.
        geos_data = GEOSGeometry.from_gml(tostring(element))
        geos_data.srid = srs.srid

        # Using the WFS 2 format (urn:ogc:def:crs:EPSG::4326"), coordinates should be latitude/longitude.
        # However, when providing legacy formats like srsName="EPSG:4326",
        # input is assumed to be in legacy longitude/latitude axis ordering too.
        # This reflects what GeoServer does: https://docs.geoserver.org/main/en/user/services/wfs/axis_order.html
        if not srs.force_xy:
            CRS.tag_geometry(geos_data, axis_order=AxisOrder.AUTHORITY)
        return cls(srs=srs, geos_data=geos_data)

    def __repr__(self):
        # Better rendering for unit test debugging
        return f"{self.__class__.__name__}(srs={self.srs!r}, geos_data=GEOSGeometry({self.geos_data.wkt!r}))"

    @property
    def wkt(self) -> str:
        """Render the Geometry as well-known text"""
        return self.geos_data.wkt

    @property
    def json(self):
        return self.geos_data.json

    def build_rhs(self, compiler: CompiledQuery):
        # Perform final validation during the construction of the query.
        if self.srs is None:
            # When the feature type is known, apply its default CRS.
            # This is not possible in XML parsing, but may happen for BBOX parsing.
            self.srs = compiler.feature_types[0].crs
            self.geos_data.srid = self.srs.srid  # assign default CRS to geometry
        elif compiler.feature_types:  # for unit tests
            self.srs = compiler.feature_types[0].resolve_crs(self.srs, locator="bbox")

        # Make sure the data is suitable for processing by the ORM.
        # The database needs the geometry in traditional (x/y) ordering.
        if self.srs.is_north_east_order:
            return self.srs.apply_to(self.geos_data, clone=True, axis_order=AxisOrder.TRADITIONAL)
        else:
            return self.geos_data


@tag_registry.register("TimeInstant", hidden=True)
@tag_registry.register("TimePeriod", hidden=True)
class AbstractTimeGeometricPrimitive(AbstractTimePrimitive):
    """Not implemented: the whole GML temporal logic.

    Examples for GML time elements include::

      <gml:TimeInstant gml:id="TI1">
         <gml:timePosition>2005-05-19T09:28:40Z</gml:timePosition>
      </gml:TimeInstant>

    and::

      <gml:TimePeriod gml:id="TP1">
         <gml:begin>
            <gml:TimeInstant gml:id="TI1">
               <gml:timePosition>2005-05-17T00:00:00Z</gml:timePosition>
            </gml:TimeInstant>
         </gml:begin>
         <gml:end>
            <gml:TimeInstant gml:id="TI2">
               <gml:timePosition>2005-05-23T00:00:00Z</gml:timePosition>
            </gml:TimeInstant>
         </gml:end>
      </gml:TimePeriod>
    """

    xml_ns = xmlns.gml32


@tag_registry.register("TimeNode", hidden=True)
@tag_registry.register("TimeEdge", hidden=True)
class AbstractTimeTopologyPrimitiveType(AbstractTimePrimitive):
    """Not implemented: GML temporal logic for TimeNode/TimeEdge."""

    xml_ns = xmlns.gml32



"""Compatability imports"""

import sys

import django
from django.conf import settings
from django.db import models

if (
    "django.contrib.postgres" in settings.INSTALLED_APPS
    or "django.contrib.postgres" in sys.modules
):
    from django.contrib.postgres.fields import ArrayField
else:
    ArrayField = None

GeneratedField = models.GeneratedField if django.VERSION >= (5, 0) else None


__all__ = (
    "ArrayField",
    "GeneratedField",
)
